# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""IUT data structure module."""
import os
import logging
from collections import OrderedDict
from packageurl import PackageURL
from jsontas.jsontas import JsonTas


class Iut:  # pylint: disable=too-few-public-methods
    """Data object for IUTs."""

    logger = logging.getLogger(__name__)

    def __init__(self, product):
        """Initialize.

        :param product: Dictionary to set attributes from.
                        Should be the response from pool plugin list.
        :type product: dict
        """
        self.test_runner = {}
        self.steps = {"environment": self.load_environment}

        product["identity"] = PackageURL.from_string(product["identity"])
        for key, value in product.items():
            setattr(self, key, value)
        self._product_dict = product
        self.jsontas = JsonTas()
        self.jsontas.dataset.add("iut", self._product_dict)
        self.prepare()

    def prepare(self):
        """Prepare IUT for testing."""
        self.logger.info("Preparing IUT %r", self)
        for step, definition in self.test_runner.get("steps", {}).items():
            step_method = self.steps.get(step)
            if step_method is None:
                self.logger.error(
                    "Step %r does not exist. Available %r", step, self.steps
                )
                continue
            self.logger.info("Executing step %r", step)
            definition = OrderedDict(**definition)
            step_result = self.jsontas.run(json_data=definition)
            step_method(step_result)

    @staticmethod
    def load_environment(environment):
        """Load and set environment variables from IUT definition.

        :param environment: Environment variables to set.
        :type environment: dict
        """
        for key, value in environment.items():
            os.environ[key] = value

    @property
    def as_dict(self):
        """Return IUT as a dictionary."""
        return self._product_dict

    def __repr__(self):
        """Represent IUT as string."""
        try:
            return self._product_dict.get("identity").to_string()
        except:  # noqa pylint:disable=bare-except
            return "Unknown"



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests full executions."""
import os
import logging
from copy import deepcopy
from shutil import rmtree
from contextlib import contextmanager
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch
from etos_lib.lib.debug import Debug
from etos_test_runner.etr import ETR

# pylint:disable=too-many-locals

SUITE = {
    "name": "Test ETOS API scenario",
    "priority": 1,
    "recipes": [
        {
            "constraints": [
                {"key": "ENVIRONMENT", "value": {}},
                {"key": "PARAMETERS", "value": {}},
                {"key": "COMMAND", "value": "exit 0"},
                {
                    "key": "TEST_RUNNER",
                    "value": "registry.nordix.org/eiffel/etos-python-test-runner:3.9.0",
                },
                {"key": "EXECUTE", "value": ["echo 'this is the pre-execution step'"]},
                {
                    "key": "CHECKOUT",
                    "value": ["echo 'this is the checkout step'"],
                },
            ],
            "id": "6e8d29eb-4b05-4f5e-9207-0c94438479c7",
            "testCase": {
                "id": "ETOS_API_functests",
                "tracker": "Github",
                "url": "https://github.com/eiffel-community/etos-api",
            },
        }
    ],
    "test_runner": "registry.nordix.org/eiffel/etos-python-test-runner:3.9.0",
    "iut": {
        "provider_id": "default",
        "identity": "pkg:docker/production/etos/etos-api@1.2.0",
        "type": "docker",
        "namespace": "production/etos",
        "name": "etos-api",
        "version": "1.2.0",
        "qualifiers": {},
        "subpath": None,
    },
    "artifact": "e9b0c120-8638-4c73-9b5c-e72226415ae6",
    "context": "fde87097-46bd-4916-b69f-48dbbec47936",
    "executor": {},
    "log_area": {
        "provider_id": "default",
        "livelogs": "http://localhost/livelogs",
        "type": "ARTIFACTORY",
        "upload": {
            "url": "http://localhost/logs/{context}/{folder}/{name}",
            "method": "POST",
        },
        "logs": {},
    },
}


class TestLogUpload(TestCase):
    """Test a log upload of ETR."""

    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(cls):
        """Create a debug instance."""
        cls.main_suite_id = "577381ad-8356-4939-ab77-02e7abe06699"
        cls.debug = Debug()

    def setUp(self):
        """Create patch objects and a test folder to execute from."""
        self.patchers = []
        self.original = Path.cwd()
        self.root = Path().joinpath("testfolder").absolute()
        if self.root.exists():
            rmtree(self.root)
        self.root.mkdir()
        os.chdir(self.root)
        self._patch_wait_for_request()
        self._patch_http_request()
        self._patch_test_suite_started_request()

    def _patch_wait_for_request(self):
        """Patch the ETOS library wait for request method."""
        patcher = patch("etos_lib.etos.Http.wait_for_request")
        self.patchers.append(patcher)
        self.wait_for_request = patcher.start()
        self.suite = deepcopy(SUITE)
        self.wait_for_request.return_value = [self.suite]

    def _patch_http_request(self):
        """Patch the ETOS library http request method."""
        patcher = patch("etos_lib.etos.Http.request")
        self.patchers.append(patcher)
        self.http_request = patcher.start()
        self.http_request.return_value = True

    def _patch_test_suite_started_request(self):
        """Patch the GraphQL request for test suite started."""
        patcher = patch("etos_test_runner.lib.testrunner.request_test_suite_started")
        self.patchers.append(patcher)
        self.request_test_suite_started = patcher.start()
        self.request_test_suite_started.return_value = [
            {
                "data": {"testSuiteCategories": []},
                "meta": {"id": self.main_suite_id},
            }
        ]

    def tearDown(self):
        """Clear the test folder and patchers."""
        os.chdir(self.original)
        rmtree(self.root, ignore_errors=True)
        for patcher in self.patchers:
            patcher.stop()

    @staticmethod
    @contextmanager
    def environ(add_environment):
        """Set environment variables in context and remove after.

        :param add_environment: Environment variables to add.
        :type add_environment: dict
        """
        current = {}
        for key, value in add_environment.items():
            current[key] = os.getenv(key)
            os.environ[key] = str(value)
        yield
        for key, value in current.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value

    @staticmethod
    def get_event(event_name, events):
        """Get an event by name.

        :param event_name: Event name to find.
        :type event_name: str
        :return: Event with the name.
        :rtype: :obj:`eiffellib.lib.events.EiffelBaseEvent`
        """
        try:
            return [event for event in events if event.meta.type == event_name][0]
        except IndexError:
            return None

    @staticmethod
    def get_events(event_name, events):
        """Get multiple events by name.

        :param event_name: Event name to find.
        :type event_name: str
        :return: Events with the name.
        :rtype: list
        """
        try:
            return [event for event in events if event.meta.type == event_name]
        except IndexError:
            return []

    def test_log_upload(self):
        """Test that log upload works as expected.

        Approval criteria:
            - Log upload shall be executed after each test and after each suite.
            - Logs shall be attached to TestSuiteFinished event.

        Test steps::
            1. Initialize and run ETR.
            2. Verify that upload works as expected.
            3. Verify that log url was attached to event.
        """
        environment = {
            "ETOS_DISABLE_SENDING_EVENTS": "1",
            "ETOS_DISABLE_RECEIVING_EVENTS": "1",
            "ETOS_GRAPHQL_SERVER": "http://localhost/graphql",
            "SUB_SUITE_URL": "http://localhost/download_suite",
            "HOME": self.root,  # There is something weird with tox and HOME. This fixes it.
        }
        with self.environ(environment):
            self.logger.info("STEP: Initialize and run ETR.")
            etr = ETR()
            etr.run_etr()
            sub_suite_started = self.get_event(
                "EiffelTestSuiteStartedEvent", self.debug.events_published
            )

            self.logger.info("STEP: Verify that upload works as expected.")
            filenames = [
                f"{self.suite['recipes'][0]['testCase']['id']}_test_output.log",
                "full_execution.log",
                "workspace.tar.gz",
            ]

            for method_call in self.http_request.call_args_list:
                verb, url, as_json = method_call.args
                data = method_call.kwargs.get("data")
                filename = Path(data.name)

                self.assertEqual(verb, self.suite["log_area"]["upload"]["method"])
                self.assertEqual(
                    url,
                    self.suite["log_area"]["upload"]["url"].format(
                        context=self.suite["context"],
                        name=filename.name,
                        folder=f"{self.main_suite_id}/{sub_suite_started.meta.event_id}",
                    ),
                )
                self.assertTrue(as_json)

                for path in filenames:
                    if filename.name == path:
                        filenames.remove(path)
                        break
                else:
                    raise AssertionError(f"{filename} should not have been uploaded.")
            self.assertEqual(len(filenames), 0)

            self.logger.info("STEP: Verify that log url was attached to event.")
            test_suite_finished = self.get_event(
                "EiffelTestSuiteFinishedEvent", self.debug.events_published
            )

            files = [
                f"{self.suite['recipes'][0]['testCase']['id']}_test_output.log",
                "full_execution.log",
            ]
            for filename in files:
                url = self.suite["log_area"]["upload"]["url"].format(
                    context=self.suite["context"],
                    name=filename,
                    folder=f"{self.main_suite_id}/{sub_suite_started.meta.event_id}",
                )
                for log in test_suite_finished.data.data.get("persistentLogs"):
                    if log["name"] == filename and log["uri"] == url:
                        test_suite_finished.data.data["persistentLogs"].remove(log)
                        break
                else:
                    raise AssertionError(
                        f"{filename} was not attached to TestSuiteFinished event."
                    )

    def test_log_upload_rename_file_too_long(self):
        """Test that log upload works and that it renames too long filenames.

        Approval criteria:
            - Log names shall be renamed if they are considered too long.

        Test steps::
            1. Initialize ETR.
            2. Run ETR with a suite that creates a file with a too long name.
            3. Verify that the file was renamed before upload.
        """
        environment = {
            "ETOS_DISABLE_SENDING_EVENTS": "1",
            "ETOS_DISABLE_RECEIVING_EVENTS": "1",
            "ETOS_GRAPHQL_SERVER": "http://localhost/graphql",
            "SUB_SUITE_URL": "http://localhost/download_suite",
            "HOME": self.root,  # There is something weird with tox and HOME. This fixes it.
        }
        suite = deepcopy(SUITE)
        command_index = suite["recipes"][0]["constraints"].index(
            {"key": "COMMAND", "value": "exit 0"}
        )
        suite["recipes"][0]["constraints"][command_index] = {
            "key": "COMMAND",
            "value": "touch $LOG_PATH/%s.txt" % ("a" * 251),
        }
        # The file that is created will total to '255' characters, the maximum limit, but since
        # ETR attaches the test case name for each log file generated by that test case we will
        # get a filename that is too long (ETOS_API_Functests_(a*251).txt) and thus should
        # be renamed.
        with self.environ(environment):
            self.logger.info("STEP: Initialize ETR.")
            etr = ETR()

            self.logger.info(
                "STEP: Run ETR with a suite that creates a file with a too long name."
            )
            self.wait_for_request.return_value = [suite]
            etr.run_etr()
            test_suite_finished = self.get_event(
                "EiffelTestSuiteFinishedEvent", self.debug.events_published
            )

            self.logger.info("STEP: Verify that the file was renamed before upload")
            for log in test_suite_finished.data.data.get("persistentLogs"):
                if log["name"].endswith("aa.txt"):
                    self.assertLessEqual(len(log["name"]), 255)
                    break
            else:
                raise AssertionError("Log file was not uploaded")

    def test_log_upload_rename_file_exists(self):
        """Test that log upload works and that it renames files with the same name.

        Approval criteria:
            - Log names shall be renamed if are named the same.

        Test steps::
            1. Initialize ETR.
            2. Run ETR with a test suite that creates two files with the same name.
            3. Verify that one file was renamed before upload.
        """
        environment = {
            "ETOS_DISABLE_SENDING_EVENTS": "1",
            "ETOS_DISABLE_RECEIVING_EVENTS": "1",
            "ETOS_GRAPHQL_SERVER": "http://localhost/graphql",
            "SUB_SUITE_URL": "http://localhost/download_suite",
            "HOME": self.root,  # There is something weird with tox and HOME. This fixes it.
        }
        command_index = self.suite["recipes"][0]["constraints"].index(
            {"key": "COMMAND", "value": "exit 0"}
        )
        self.suite["recipes"][0]["constraints"][command_index] = {
            "key": "COMMAND",
            "value": "touch $LOG_PATH/my_file.txt",
        }
        recipe = deepcopy(self.suite["recipes"][0])
        self.suite["recipes"].append(recipe)
        # Same test has been added twice. Both tests creates file "my_file.txt"
        # and the tests have the same name.
        # Filenames will be "ETOS_API_Functest_my_file.txt"
        with self.environ(environment):
            self.logger.info("STEP: Initialize ETR.")
            etr = ETR()

            self.logger.info(
                "STEP: Run ETR with a test suite that creates two files with the same name."
            )
            etr.run_etr()

            self.logger.info("STEP: Verify that the file was renamed before upload")
            sub_suite_started = self.get_event(
                "EiffelTestSuiteStartedEvent", self.debug.events_published
            )
            test_suite_finished = self.get_event(
                "EiffelTestSuiteFinishedEvent", self.debug.events_published
            )
            expected = [
                {
                    "name": "ETOS_API_functests_my_file.txt",
                    "uri": self.suite["log_area"]["upload"]["url"].format(
                        context=self.suite["context"],
                        folder=f"{self.main_suite_id}/{sub_suite_started.meta.event_id}",
                        name="ETOS_API_functests_my_file.txt",
                    ),
                },
                {
                    "name": "1_ETOS_API_functests_my_file.txt",
                    "uri": self.suite["log_area"]["upload"]["url"].format(
                        context=self.suite["context"],
                        folder=f"{self.main_suite_id}/{sub_suite_started.meta.event_id}",
                        name="1_ETOS_API_functests_my_file.txt",
                    ),
                },
            ]
            for log in test_suite_finished.data.data.get("persistentLogs"):
                if "ETOS_API_functests_my_file.txt" in log["name"]:
                    self.assertIn(log, expected)
                    expected.remove(log)
            self.assertEqual(len(expected), 0)

    def test_artifact_upload(self):
        """Test that artifact upload works as expected.

        Approval criteria:
            - Artifact upload shall be executed after each test and after each suite.
            - Artifacts shall be attached to ArtifactCreated event.

        Test steps::
            1. Initialize and run ETR suite which creates an artifact.
            2. Verify that upload works as expected.
            3. Verify that artifact was attached to event.
        """
        environment = {
            "ETOS_DISABLE_SENDING_EVENTS": "1",
            "ETOS_DISABLE_RECEIVING_EVENTS": "1",
            "ETOS_GRAPHQL_SERVER": "http://localhost/graphql",
            "SUB_SUITE_URL": "http://localhost/download_suite",
            "HOME": self.root,  # There is something weird with tox and HOME. This fixes it.
        }
        command_index = self.suite["recipes"][0]["constraints"].index(
            {"key": "COMMAND", "value": "exit 0"}
        )
        self.suite["recipes"][0]["constraints"][command_index] = {
            "key": "COMMAND",
            "value": "touch $ARTIFACT_PATH/my_artifact.txt",
        }
        with self.environ(environment):
            self.logger.info(
                "STEP: Initialize and run ETR suite which creates an artifact."
            )
            etr = ETR()
            etr.run_etr()
            self.logger.info("STEP: Verify that upload works as expected.")
            sub_suite_started = self.get_event(
                "EiffelTestSuiteStartedEvent", self.debug.events_published
            )
            test_name = self.suite["recipes"][0]["testCase"]["id"]
            suite_name = self.suite.get("name").replace(" ", "-")
            artifact = f"{test_name}_my_artifact.txt"

            # We only verify that test artifact upload works. Global artifact is tested by
            # test_log_upload.
            for method_call in self.http_request.call_args_list:
                verb, url, as_json = method_call.args
                data = method_call.kwargs.get("data")
                filename = Path(data.name)
                if filename.name == artifact:
                    self.assertEqual(verb, self.suite["log_area"]["upload"]["method"])
                    self.assertEqual(
                        url,
                        self.suite["log_area"]["upload"]["url"].format(
                            context=self.suite["context"],
                            name=filename.name,
                            folder=f"{self.main_suite_id}/{sub_suite_started.meta.event_id}",
                        ),
                    )
                    self.assertTrue(as_json)
                    break
            else:
                raise AssertionError(f"Artifact {artifact} was not uploaded.")

            self.logger.info("STEP: Verify that artifact was attached to event.")
            artifact_created = self.get_events(
                "EiffelArtifactCreatedEvent", self.debug.events_published
            )
            self.assertEqual(len(artifact_created), 2)
            test_created = artifact_created.pop(0)
            suite_created = artifact_created.pop(0)
            self.assertEqual(
                test_created.data.data.get("identity"),
                f"pkg:etos-test-output/{suite_name}/{test_name}",
            )
            self.assertListEqual(
                test_created.data.data.get("fileInformation"), [{"name": artifact}]
            )
            self.assertEqual(
                suite_created.data.data.get("identity"),
                f"pkg:etos-test-output/{suite_name}",
            )
            self.assertListEqual(
                suite_created.data.data.get("fileInformation"),
                [{"name": "workspace.tar.gz"}],
            )
            self.assertEqual(len(artifact_created), 0)
            artifact_published = self.get_events(
                "EiffelArtifactPublishedEvent", self.debug.events_published
            )
            self.assertEqual(len(artifact_published), 2)
            my_artifact = artifact_published.pop(0)
            url = self.suite["log_area"]["upload"]["url"].format(
                context=self.suite["context"],
                name="",
                folder=f"{self.main_suite_id}/{sub_suite_started.meta.event_id}",
            )
            self.assertListEqual(
                my_artifact.data.data.get("locations"),
                [{"type": "ARTIFACTORY", "uri": url}],
            )

            url = self.suite["log_area"]["upload"]["url"].format(
                context=self.suite["context"],
                name="",
                folder=f"{self.main_suite_id}/{sub_suite_started.meta.event_id}",
            )
            workspace = artifact_published.pop(0)
            self.assertListEqual(
                workspace.data.data.get("locations"),
                [{"type": "ARTIFACTORY", "uri": url}],
            )
            self.assertEqual(len(artifact_published), 0)



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETR executor module."""
import os
import shlex
import logging
import signal
import json
import re
from pathlib import Path
from shutil import copy
from pprint import pprint

BASE = Path(__file__).parent.absolute()


class TestCheckoutTimeout(TimeoutError):
    """Test checkout timeout exception."""


def _test_checkout_signal_handler(signum, frame):  # pylint:disable=unused-argument
    """Raise timeout error on test checkout."""
    raise TestCheckoutTimeout("Took too long to checkout test cases.")


class Executor:  # pylint:disable=too-many-instance-attributes
    """Execute a single test-case, -class, -module, -folder etc."""

    report_path = "test_output.log"
    test_name = ""
    current_test = None
    test_regex = {}
    logger = logging.getLogger("Executor")

    def __init__(self, test, iut, etos):
        """Initialize.

        :param test: Test to execute.
        :type test: str
        :param iut: IUT to execute test on.
        :type iut: :obj:`etr.lib.iut.Iut`
        :param etos: ETOS library instance.
        :type etos: :obj:`etos_lib.etos.Etos`
        """
        self.load_regex()
        self.test = test
        self.tests = {}

        self.test_environment_variables = {}
        self.test_command = None
        self.pre_test_execution = []
        self.test_command_input_arguments = {}
        self.checkout_command = []

        self.constraints = test.get("constraints", [])
        for constraint in self.constraints:
            if constraint.get("key") == "ENVIRONMENT":
                self.test_environment_variables = constraint.get("value")
            elif constraint.get("key") == "COMMAND":
                self.test_command = constraint.get("value")
            elif constraint.get("key") == "EXECUTE":
                self.pre_test_execution = constraint.get("value")
            elif constraint.get("key") == "PARAMETERS":
                self.test_command_input_arguments = constraint.get("value")
            elif constraint.get("key") == "CHECKOUT":
                self.checkout_command = constraint.get("value")

        self.test_name = test.get("testCase").get("id")
        self.test_id = test.get("id")
        self.iut = iut
        self.etos = etos
        self.context = self.etos.config.get("context")
        self.result = True

    def load_regex(self):
        """Attempt to load regex file from environment variables.

        The regex file is used to determine when a test case has triggered,
        started, passed, failed, been skipped, raise error and the test name.
        """
        if os.getenv("TEST_REGEX"):
            try:
                path = Path(os.getenv("TEST_REGEX"))
                if path.exists() and path.is_file():
                    regex = json.load(path.open())
                    for key, value in regex.items():
                        self.test_regex[key] = re.compile(value)
                else:
                    self.logger.warning("%r is not a file or does not exist.", path)
            except TypeError as exception:
                self.logger.error("%r", exception)
                self.logger.error("Wrong type when loading %r", path)
            except re.error as exception:
                self.logger.error("%r", exception)
                self.logger.error("Failed to parse regex in file %r (%r)", path, value)
            except json.decoder.JSONDecodeError as exception:
                self.logger.error("%r", exception)
                self.logger.error("Failed to load JSON %r", path)
            except Exception as exception:  # pylint:disable=broad-except
                self.logger.error("%r", exception)
                self.logger.error("Unknown error when loading regex JSON file.")

    def _checkout_tests(self, test_checkout, workspace):
        """Check out tests for this execution.

        :param test_checkout: Test checkout parameters from test suite.
        :type test_checkout: list
        :param workspace: The workspace directory where the checkout script should be placed.
        :type workspace: :obj:`pathlib.Path`
        """
        test_directory_name = Path().absolute().name
        checkout = workspace.joinpath(f"checkout_{test_directory_name}.sh")
        with checkout.open(mode="w") as checkout_file:
            checkout_file.write('eval "$(pyenv init -)"\n')
            checkout_file.write("pyenv shell --unset\n")
            for command in test_checkout:
                checkout_file.write("{} || exit 1\n".format(command))

        self.logger.info("Checkout script:\n" "%s", checkout.read_text())

        signal.signal(signal.SIGALRM, _test_checkout_signal_handler)
        signal.alarm(60)
        try:
            success, output = self.etos.utils.call(
                ["/bin/bash", str(checkout)], shell=True, wait_output=False
            )
        finally:
            signal.alarm(0)
        if not success:
            pprint(output)
            raise Exception("Could not checkout tests using '{}'".format(test_checkout))

    def _build_test_command(self):
        """Build up the actual test command based on data from event."""
        base_executor = Path(BASE).joinpath("executor.sh")
        executor = Path().joinpath("executor.sh")
        copy(base_executor, executor)

        self.logger.info("Executor script:\n" "%s", executor.read_text())

        test_command = ""
        parameters = []

        for parameter, value in self.test_command_input_arguments.items():
            if value == "":
                parameters.append(parameter)
            else:
                parameters.append("{}={}".format(parameter, value))

        test_command = "./{} {} {} 2>&1".format(
            str(executor), self.test_command, " ".join(parameters)
        )
        return test_command

    def __enter__(self):
        """Enter context and set current test."""
        self.etos.config.set("current_test", self.test_name)
        return self

    def __exit__(self, _type, value, traceback):
        """Exit context and unset current test."""
        self.etos.config.set("current_test", None)

    def _pre_execution(self, command):
        """Write pre execution command to a shell script.

        :param command: Environment and pre execution shell command to write to shell script.
        :type command: str
        """
        environ = Path().joinpath("environ.sh")
        with environ.open(mode="w") as environ_file:
            for arg in command:
                environ_file.write("{} || exit 1\n".format(arg))
        self.logger.info(
            "Pre-execution script (includes ENVIRONMENT):\n" "%s",
            environ.read_text(),
        )

    def _build_environment_command(self):
        """Build command for setting environment variables prior to execution.

        :return: Command to run pre execution.
        :rtype: str
        """
        environments = [
            "export {}={}".format(key, shlex.quote(value))
            for key, value in self.test_environment_variables.items()
        ]
        return environments + self.pre_test_execution

    def _triggered(self, test_name):
        """Send a test case triggered event.

        :param test_name: Name of test that is triggered.
        :type test_name: str
        :return: Test case triggered event created and sent.
        :rtype: :obj:`eiffellib.events.eiffel_test_case_triggered_event.EiffelTestCaseTriggeredEvent`  # pylint:disable=line-too-long
        """
        return self.etos.events.send_test_case_triggered(
            {"id": test_name},
            self.etos.config.get("artifact"),
            links={"CONTEXT": self.context},
        )

    def _started(self, test_name):
        """Send a testcase started event.

        :param test_name: Name of test that has started.
        :type test_name: str
        :return: Test case started event created and sent.
        :rtype: :obj:`eiffellib.events.eiffel_test_case_started_event.EiffelTestCaseStartedEvent`
        """
        triggered = self.tests[test_name].get("triggered")
        if triggered is None:
            return None
        return self.etos.events.send_test_case_started(
            triggered, links={"CONTEXT": self.context}
        )

    def _finished(self, test_name, result):
        """Send a testcase finished event.

        :param test_name: Name of test that is finished.
        :type test_name: str
        :param result: Result of test case.
        :type result: str
        :return: Test case finished event created and sent.
        :rtype: :obj:`eiffellib.events.eiffel_test_case_finished_event.EiffelTestCaseFinishedEvent`
        """
        triggered = self.tests[test_name].get("triggered")
        if triggered is None:
            return None

        if result == "ERROR":
            outcome = {"verdict": "FAILED", "conclusion": "INCONCLUSIVE"}
        elif result == "FAILED":
            outcome = {"verdict": "FAILED", "conclusion": "FAILED"}
        elif result == "SKIPPED":
            outcome = {
                "verdict": "PASSED",
                "conclusion": "SUCCESSFUL",
                "description": "SKIPPED",
            }
        else:
            outcome = {"verdict": "PASSED", "conclusion": "SUCCESSFUL"}
        return self.etos.events.send_test_case_finished(
            triggered, outcome, links={"CONTEXT": self.context}
        )

    def parse(self, line):
        """Parse test output in order to send test case events.

        :param line: Line to parse.
        :type line: str
        """
        if not isinstance(line, str):
            return
        test_name = self.test_regex["test_name"].findall(line)
        if test_name:
            self.current_test = test_name[0]
            self.tests.setdefault(self.current_test, {})
        if self.test_regex["triggered"].match(line):
            self.tests[self.current_test]["triggered"] = self._triggered(
                self.current_test
            )
        if self.test_regex["started"].match(line):
            self.tests[self.current_test]["started"] = self._started(self.current_test)
        if self.test_regex["passed"].match(line):
            self.tests[self.current_test]["finished"] = self._finished(
                self.current_test, "PASSED"
            )
        if self.test_regex["failed"].match(line):
            self.tests[self.current_test]["finished"] = self._finished(
                self.current_test, "FAILED"
            )
        if self.test_regex["error"].match(line):
            self.tests[self.current_test]["finished"] = self._finished(
                self.current_test, "ERROR"
            )
        if self.test_regex["skipped"].match(line):
            self.tests[self.current_test]["finished"] = self._finished(
                self.current_test, "SKIPPED"
            )

    def execute(self, workspace):
        """Execute a test case.

        :param workspace: Workspace instance for creating test directories.
        :type workspace: :obj:`etos_test_runner.lib.workspace.Workspace`
        """
        line = False
        with workspace.test_directory(
            " ".join(self.checkout_command),
            self._checkout_tests,
            self.checkout_command,
            workspace.workspace,
        ) as test_directory:
            self.report_path = test_directory.joinpath(f"logs/{self.report_path}")
            self.logger.info("Report path: %r", self.report_path)

            self.logger.info("Build pre-execution script.")
            self._pre_execution(self._build_environment_command())

            self.logger.info("Build test command")
            command = self._build_test_command()

            self.logger.info("Run test command: %r", command)
            iterator = self.etos.utils.iterable_call(
                [command], shell=True, executable="/bin/bash", output=self.report_path
            )

            self.logger.info("Wait for test to finish.")
            for _, line in iterator:
                if self.test_regex:
                    self.parse(line)
            self.result = line
            self.logger.info("Finished with result %r.", self.result)



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETR workspace module."""
from os import chdir, environ
import logging
from contextlib import contextmanager
from tempfile import mkdtemp
from pathlib import Path
from shutil import make_archive


class Workspace:
    """Test runner workspace. This is where all testing is done."""

    logger = logging.getLogger(__name__)
    workspace = None
    compressed_workspace = None

    def __init__(self, log_area):
        """Initialize workspace with a dictionary of identifiers.

        :param log_area: Log area library for collecting and uploading logs.
        :type log_area: :obj:`etos_test_runner.lib.log_area.LogArea`
        """
        self.log_area = log_area

        self.top_dir = Path.cwd()
        self.identifiers = {}

        self.global_logs = self.top_dir.joinpath("logs")
        environ["GLOBAL_LOGS"] = str(self.global_logs)

    def __enter__(self, path=None):
        """Create and chdir to a workspace."""
        self.workspace = Path(path) if path else self.top_dir.joinpath("workspace")
        self.logger.info(
            "Creating workspace directory %r", self.workspace.relative_to(self.top_dir)
        )
        self.workspace.mkdir(exist_ok=True)
        self.global_logs.mkdir(exist_ok=True)
        self.logger.info(
            "Changing directory to workspace %r",
            self.workspace.relative_to(self.top_dir),
        )
        chdir(self.workspace)
        return self

    def __exit__(self, _type, _value, _traceback):
        """Compress and cleanup workspace."""
        self.logger.info("Returning to %r", self.top_dir)
        chdir(self.top_dir)
        self.compress()
        self.log_area.upload_logs(self.log_area.collect(self.global_logs))
        self.log_area.upload_artifacts(
            [
                {
                    "name": self.compressed_workspace.name,
                    "file": self.compressed_workspace,
                }
            ]
        )

    def add_to_full_log(self, report):
        """Add text in report to the full global log.

        :param report: Path the the test report generated by test case.
        :type report: :obj:`pathlib.Path`
        """
        if report.exists():
            full_execution = self.global_logs.joinpath("full_execution.log")
            full_execution.touch()
            with full_execution.open(mode="a") as full_execution_log:
                full_execution_log.write(report.read_text())

    @contextmanager
    def collect_logs(self, base_path):
        """Create log and artifact path and collect them after context.

        :param base_path: Base starting path for logs.
        :type base_path: :obj:`pathlib.Path`
        """
        log_path = base_path.joinpath("logs")
        artifact_path = base_path.joinpath("artifacts")
        log_path.mkdir(exist_ok=True)
        artifact_path.mkdir(exist_ok=True)
        # DEPRECATED: TEST_ARTIFACT_PATH is deprecated and only exists
        # for backwards compatability.
        environ["TEST_ARTIFACT_PATH"] = str(artifact_path)
        environ["ARTIFACT_PATH"] = str(artifact_path)
        environ["LOG_PATH"] = str(log_path)
        yield
        del environ["TEST_ARTIFACT_PATH"]
        del environ["ARTIFACT_PATH"]
        del environ["LOG_PATH"]
        self.add_to_full_log(log_path.joinpath("test_output.log"))
        self.log_area.upload_logs(self.log_area.collect(log_path))
        self.log_area.upload_artifacts(self.log_area.collect(artifact_path))

    @contextmanager
    def test_directory(
        self, identifier, on_create=None, *on_create_args
    ):  # pylint:disable=keyword-arg-before-vararg
        """Create and chdir to a subfolder within the workspace.

        :param identifier: Identifier for this new test_directory.
                           Does not create a new directory if directory with
                           identifier already exists.
        :type identifier: str
        :param on_create: Call callable on creation of directory.
        :type on_create: callable
        :param on_create_args: Positional arguments for on_create callable.
        :type on_create_args: tuple
        :return: Path to the newly created directory.
        :rtype: :obj:`pathlib.Path`
        """
        if self.workspace is None or not self.workspace.is_dir():
            raise Exception("Workspace not created.")
        try:
            self.logger.info("Getting test directory with identifier %r", identifier)
            if self.identifiers.get(identifier) is None:
                directory = Path(mkdtemp(dir=self.workspace))
                self.logger.info(
                    "Directory not found. Created %r",
                    directory.relative_to(self.top_dir),
                )
                self.identifiers[identifier] = directory
                self.logger.info(
                    "Change directory to %r",
                    self.identifiers.get(identifier).relative_to(self.top_dir),
                )
                chdir(self.identifiers.get(identifier))
                if on_create is not None:
                    on_create(*on_create_args)
            else:
                self.logger.info(
                    "Found %r",
                    self.identifiers.get(identifier).relative_to(self.top_dir),
                )
                self.logger.info(
                    "Change directory to %r",
                    self.identifiers.get(identifier).relative_to(self.top_dir),
                )
                chdir(self.identifiers.get(identifier))
            with self.collect_logs(self.identifiers.get(identifier)):
                yield self.identifiers.get(identifier)
        finally:
            self.logger.info(
                "Returning to %r", self.workspace.relative_to(self.top_dir)
            )
            chdir(self.workspace)

    def compress(self):
        """Compress the entire workspace folder."""
        if self.workspace is None or not self.workspace.is_dir():
            raise Exception("Workspace not created.")
        self.logger.info("Compress workspace directory")
        compressed_workspace = self.top_dir.joinpath("workspace").relative_to(
            Path.cwd()
        )
        filename = make_archive(
            compressed_workspace,
            format="gztar",
            root_dir=self.top_dir.relative_to(Path.cwd()),
            base_dir=self.workspace.relative_to(Path.cwd()),
            logger=self.logger,
        )
        self.compressed_workspace = Path(filename)



#!/usr/bin/env python
# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-
"""conftest.py for etos_test_runner.

If you don't know what this is for, just leave it empty.
Read more about conftest.py under:
https://pytest.org/latest/plugins.html
"""
import pytest
from etos_lib.lib.debug import Debug
from etos_lib.lib.config import Config


@pytest.fixture(scope="function", autouse=True)
def clear_etos_lib_configurations():
    """Make sure that etos library configuration is cleared after each test."""
    config = Config()
    debug = Debug()
    yield
    config.reset()
    debug.events_received.clear()
    debug.events_published.clear()



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETR log area handler."""
import logging
import traceback
import time
from copy import deepcopy
from pathlib import Path
from shutil import make_archive, rmtree
from json.decoder import JSONDecodeError

from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from requests.exceptions import HTTPError
from urllib3.exceptions import MaxRetryError, NewConnectionError


class LogArea:
    """Library for uploading logs to log area."""

    logger = logging.getLogger(__name__)

    def __init__(self, etos):
        """Initialize with an ETOS instance.

        :param etos: Instance of ETOS library.
        :type etos: :obj:`etos_lib.etos.Etos`
        """
        self.etos = etos
        self.suite_name = (
            self.etos.config.get("test_config").get("name").replace(" ", "-")
        )
        self.log_area = self.etos.config.get("test_config").get("log_area")
        self.logs = []
        self.artifacts = []

    @property
    def persistent_logs(self):
        """All persistent log formatted for EiffelTestSuiteFinishedEvent.

        :return: All persistent logs.
        :rtype: list
        """
        return [{"name": log["name"], "uri": log["uri"]} for log in self.logs]

    def _fix_name(self, path, test_name=None):
        """Fix the name of a file.

        "Fixing" means:

            - Prepend test name if file was gathered by a test case.
            - Prepend "log_info" if supplied by environment provider.
            - Reduce the size of the filename if it's too large.
            - Prepend a counter if filename already exists.

        :param path: Path to a file to fix.
        :type path: :obj:`pathlib.Path`
        :param test_name: Test name if this file was gathered during a test case.
        :type test_name: str
        :return: A new and improved path and filename.
        :rtype: :obj:`pathlib.Path`
        """
        directory, filename = path.parent, path.name
        if test_name is not None:
            self.logger.info(
                "File collected as part of test case. Prepending %r", test_name
            )
            filename = f"{test_name}_{filename}"
            self.logger.info("Result: %r", filename)
        if self.log_area.get("logs"):
            prepend = self.log_area.get("logs").get("prepend", "")
            join_character = self.log_area.get("logs").get("join_character", "_")
            self.logger.info(
                "Log instructions added by environment provider. Prepending %r", prepend
            )
            filename = f"{prepend}{join_character}{filename}"
            self.logger.info("Result: %r", filename)
        if len(filename) + 5 > 255:  # +5 as to be able to prepend counter.
            self.logger.info(
                "Filename is too long at %r. Reduce size to %r.", len(filename) + 5, 250
            )
            filename = filename[: 250 - len(path.suffix)] + path.suffix
            self.logger.info("Result: %r", filename)
        log_names = [item["name"] for item in self.logs + self.artifacts]
        index = 0
        while filename in log_names:
            index += 1
            filename = f"{index}_{filename}"
            self.logger.info("Log name already exists. Rename to %r.", filename)
        return path.rename(directory.joinpath(filename))

    def collect(self, path):
        """Collect logs and artifacts from path.

        :param path: Path to collect logs and artifacts from.
        :type path: :obj:`pathlib.Path`
        :return: Filenames and paths.
        :rtype: list
        """
        test_name = self.etos.config.get("current_test")
        items = []
        self.logger.info("Collecting logs/artifacts for %r", test_name or "global")
        for item in path.iterdir():
            if item.is_dir():
                compressed_item = make_archive(
                    item.relative_to(Path.cwd()),
                    format="gztar",
                    root_dir=path,
                    base_dir=item.name,
                    logger=self.logger,
                )
                rmtree(item)
                item = Path(compressed_item)
            item = self._fix_name(item, test_name)
            items.append({"name": item.name, "file": item})
        return items

    def upload_logs(self, logs):
        """Upload logs to log area.

        :param logs: Logs to upload.
        :type logs: list
        """
        log_area_folder = "{}/{}".format(
            self.etos.config.get("main_suite_id"), self.etos.config.get("sub_suite_id")
        )
        for log in logs:
            log["uri"] = self.__upload(
                self.etos.config.get("context"),
                log["file"],
                log["name"],
                log_area_folder,
            )
            self.logs.append(log)
            log["file"].unlink()

    def _artifact_created(self, artifacts):
        """Send artifact created event.

        :param artifacts: Artifacts that exists within this event.
        :type artifacts: list
        """
        test_name = self.etos.config.get("current_test")
        if test_name:
            identity = f"pkg:etos-test-output/{self.suite_name}/{test_name}"
        else:
            identity = f"pkg:etos-test-output/{self.suite_name}"
        file_information = []
        for artifact in artifacts:
            file_information.append({"name": artifact["file"].name})
        return self.etos.events.send_artifact_created_event(
            identity,
            fileInformation=file_information,
            links={
                "CONTEXT": self.etos.config.get("context"),
                "CAUSE": self.etos.config.get("sub_suite_id"),
            },
        )

    def _artifact_published(self, artifact_created, published_url):
        """Send artifact published event.

        :param artifact_created: The created artifact to publish.
        :type artifact_created: :obj:`eiffellib.events.EiffelArtifactCreatedEvent`
        :param published_url: URL to the published directory.
        :type published_url: str
        """
        log_area_type = self.log_area.get("type", "OTHER")
        locations = [{"uri": published_url, "type": log_area_type}]
        self.etos.events.send_artifact_published_event(
            locations,
            artifact_created,
            links={"CONTEXT": self.etos.config.get("context")},
        )

    def upload_artifacts(self, artifacts):
        """Upload artifacts to log area.

        :param artifacts: Artifacs to upload.
        :type artifacts: list
        :return: Artifact name and URI
        :rtype: tuple
        """
        if not artifacts:
            return
        log_area_folder = "{}/{}".format(
            self.etos.config.get("main_suite_id"), self.etos.config.get("sub_suite_id")
        )
        self.logger.info("Uploading artifacts %r to log area", artifacts)
        artifact_created = self._artifact_created(artifacts)

        for artifact in artifacts:
            artifact["uri"] = self.__upload(
                self.etos.config.get("context"),
                artifact["file"],
                artifact["name"],
                log_area_folder,
            )
            self.artifacts.append(artifact)
            artifact["file"].unlink()

        upload = deepcopy(self.log_area.get("upload"))
        data = {
            "context": self.etos.config.get("context"),
            "folder": log_area_folder,
            "name": "",
        }
        published_url = upload["url"].format(**data)
        self._artifact_published(artifact_created, published_url)

    def __upload(self, context, log, name, folder):
        """Upload log to a storage location.

        :param context: Context for the http request.
        :type context: str
        :param log: Path to the log to upload.
        :type log: str
        :param name: Name of file to upload.
        :type name: str
        :param folder: Folder to upload to.
        :type folder: str
        :return: URI where log was uploaded to.
        :rtype: str
        """
        upload = deepcopy(self.log_area.get("upload"))
        data = {"context": context, "name": name, "folder": folder}

        # ETOS Library, for some reason, uses the key 'verb' instead of 'method'
        # for HTTP method.
        upload["verb"] = upload.pop("method")
        upload["url"] = upload["url"].format(**data)
        upload["timeout"] = upload.get("timeout", 30)
        if upload.get("auth"):
            upload["auth"] = self.__auth(**upload["auth"])

        with open(log, "rb") as log_file:
            for _ in range(3):
                request_generator = self.__retry_upload(log_file=log_file, **upload)
                try:
                    for response in request_generator:
                        self.logger.debug("%r", response)
                        if not upload.get("as_json", True):
                            self.logger.debug("%r", response.text)
                        self.logger.info("Uploaded log %r.", log)
                        self.logger.info("Upload URI          %r", upload["url"])
                        self.logger.info("Data:               %r", data)
                        break
                    break
                except:  # noqa pylint:disable=bare-except
                    self.logger.error("%r", traceback.format_exc())
                    self.logger.error("Failed to upload log!")
                    self.logger.error("Attempted upload of %r", log)
        return upload["url"]

    def __retry_upload(
        self, verb, url, log_file, timeout=None, as_json=True, **requests_kwargs
    ):  # pylint:disable=too-many-arguments
        """Attempt to connect to url for x time.

        :param verb: Which HTTP verb to use. GET, PUT, POST
                     (DELETE omitted)
        :type verb: str
        :param url: URL to retry upload request
        :type url: str
        :param log_file: Opened log file to upload.
        :type log_file: file
        :param timeout: How long, in seconds, to retry request.
        :type timeout: int or None
        :param as_json: Whether or not to return json instead of response.
        :type as_json: bool
        :param request_kwargs: Keyword arguments for the requests command.
        :type request_kwargs: dict
        :return: HTTP response or json.
        :rtype: Response or dict
        """
        if timeout is None:
            timeout = self.etos.debug.default_http_timeout
        end_time = time.time() + timeout
        self.logger.debug(
            "Retrying URL %s for %d seconds with a %s request.", url, timeout, verb
        )
        iteration = 0
        while time.time() < end_time:
            iteration += 1
            self.logger.debug("Iteration: %d", iteration)
            try:
                # Seek back to the start of the file so that the uploaded file
                # is not 0 bytes in size.
                log_file.seek(0)
                yield self.etos.http.request(
                    verb, url, as_json, data=log_file, **requests_kwargs
                )
                break
            except (
                ConnectionError,
                HTTPError,
                NewConnectionError,
                MaxRetryError,
                TimeoutError,
                JSONDecodeError,
            ):
                self.logger.warning("%r", traceback.format_exc())
                time.sleep(2)
        else:
            raise ConnectionError(
                "Unable to {} {} with params {}".format(verb, url, requests_kwargs)
            )

    @staticmethod
    def __auth(username, password, type="basic"):  # pylint:disable=redefined-builtin
        """Create an authentication for HTTP request.

        :param username: Username to authenticate.
        :type username: str
        :param password: Password to authenticate with.
        :type password: str
        :param type: Type of authentication. 'basic' or 'digest'.
        :type type: str
        :return: Authentication method.
        :rtype: :obj:`requests.auth`
        """
        if type.lower() == "basic":
            return HTTPBasicAuth(username, password)
        return HTTPDigestAuth(username, password)



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Workspace test module."""
import sys
import logging
from shutil import rmtree, unpack_archive
from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock
from etos_test_runner.lib.workspace import Workspace

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)


class TestWorkspace(TestCase):
    """Tests for the workspace library."""

    workspace = None
    logger = logging.getLogger(__name__)

    def setUp(self):
        self.items = []

    def tearDown(self):
        """Attempt to delete all created folders and files."""
        if self.workspace is not None:
            rmtree(self.workspace.workspace, ignore_errors=True)
            Path.cwd().joinpath("workspace.tar.gz").unlink(missing_ok=True)
        for item in self.items:
            if item.is_dir():
                rmtree(item)
            elif item.is_file():
                item.unlink()
        self.workspace = None

    def test_init_does_not_create(self):
        """Test that the library does not create workspace when created normally.

        Approval criteria:
            - Workspace library shall not create the workspace main directory
              when created normally.

        Test steps::
            1. Initialize the workspace library.
            2. Verify that the workspace library did not create a workspace directory.
        """
        self.logger.info("STEP: Initialize the workspace library.")
        self.workspace = Workspace(Mock())

        self.logger.info(
            "STEP: Verify that the workspace library did not create a workspace "
            "directory."
        )
        self.assertFalse(Path.cwd().joinpath("workspace").exists())

    def test_context(self):
        """Test that the library creates a workspace when launched as context.

        Approval criteria:
            - The workspace library shall create the workspace main directory when used as context
            - The workspace library shall attempt to compress the workspace directory when closed

        Test steps::
            1. Initialize the workspace library as a context manager.
            2. Verify that the workspace library created the main workspace directory.
            3. Exit the context.
            4. Verify that the workspace library compressed the main workspace directory.
        """
        self.logger.info("STEP: Initialize the workspace library as a context manager.")
        with Workspace(Mock()) as workspace:
            self.workspace = workspace
            self.logger.info(
                "STEP: Verify that the workspace library created the main "
                "workspace directory."
            )
            self.assertTrue(
                workspace.workspace.exists() and workspace.workspace.is_dir()
            )

            self.logger.info("STEP: Exit the context.")
        self.logger.info(
            "STEP: Verify that the workspace library compressed the main "
            "workspace directory."
        )
        compressed_workspace = Path.cwd().joinpath("workspace.tar.gz")
        self.assertTrue(
            compressed_workspace.exists() and compressed_workspace.is_file()
        )

    def test_compress(self):
        """Test that the compression method compresses a folder and all its subdirectories.

        Approval criteria:
            - The compression method shall compress a folder with gztar.

        Test steps::
            1. Create the workspace directory to be compressed.
            2. Compress the directory.
            3. Verify that the directory was compressed using the gztar format.
        """
        self.logger.info("STEP: Create the workspace directory to be compressed.")
        workspace = Workspace(Mock)
        directory = Path.cwd().joinpath("workspace")
        directory.mkdir()
        workspace.workspace = directory

        # Create a file to verify compression.
        directory.joinpath("file.txt").touch()

        test_folder = Path.cwd().joinpath("testfolder")
        test_folder.mkdir()
        self.items.append(test_folder)

        self.logger.info("STEP: Compress the directory.")
        workspace.compress()

        self.logger.info(
            "STEP: Verify that the directory was compressed using the gztar format."
        )
        self.items.append(test_folder)
        compressed_workspace = Path.cwd().joinpath("workspace.tar.gz")
        unpack_archive(compressed_workspace, test_folder, format="gztar")
        compressed_file = test_folder.joinpath("workspace/file.txt")
        self.assertTrue(compressed_file.exists() and compressed_file.is_file())

    def test_compress_error_when_no_workspace_exists(self):
        """Test that the compression method raises exception when no workspace exists.

        Approval criteria:
            - The compression method shall raise exception is the workspace directory does not
              exist

        Test steps::
            1. Attempt to compress a directory that does not exist.
            2. Verify that an exception was raised.
        """
        self.logger.info("STEP: Attempt to compress a directory that does not exist.")
        self.workspace = Workspace(Mock())
        with self.assertRaises(Exception):
            self.logger.info("STEP: Verify that an exception was raised.")
            self.workspace.compress()

    def test_test_directory(self):
        """Test that the workspace library can create a test directory as a context manager.

        Approval criteria:
            - The workspace library shall create and chdir to a test directory.

        Test steps::
            1. Initialize the workspace.
            2. Enter test directory in a context manager.
            3. Verify that directory was created and changed to.
        """
        self.logger.info("STEP: Initialize the workspace.")
        with Workspace(Mock()) as workspace:
            self.workspace = workspace

            self.logger.info("STEP: Enter test directory in a context manager.")
            with workspace.test_directory("dir") as directory:
                self.logger.info(
                    "STEP: Verify that directory was created and changed to."
                )
                self.assertTrue(directory.exists() and directory.is_dir())
                try:
                    self.assertEqual(Path.cwd().relative_to(directory), Path("."))
                except ValueError as exception:
                    raise AssertionError("Directory is not the cwd") from exception

    def test_test_directory_identifer_exists(self):
        """Test that the workspace does not create a directory for an identifer that already exists.

        Approval criteria:
            - The workspace library shall reuse a test directory if identifier already exists.

        Test steps::
            1. Initialize the workspace.
            2. Enter a test directory in a context manager with identifier 'dir1'.
            3. Check that test directory was created and exit the context.
            4. Enter a test directory in a context manager with the same identifer.
            5. Verify that the folder was re-used.
        """
        self.logger.info("STEP: Initialize the workspace.")
        with Workspace(Mock()) as workspace:
            self.workspace = workspace

            self.logger.info(
                "STEP: Enter a test directory in a context manager with identifier "
                "'dir1'."
            )
            with workspace.test_directory("dir1") as directory:
                self.logger.info(
                    "STEP: Check that test directory was created and exit the "
                    "context."
                )
                if not directory.exists() and directory.is_dir():
                    raise Exception("Test directory was not properly created.")
                first_stat = directory.stat()

            with workspace.test_directory("dir1") as directory:
                self.logger.info(
                    "STEP: Enter a test directory in a context manager with the "
                    "same identifer."
                )
                if not directory.exists() and directory.is_dir():
                    raise Exception("Test directory was not properly created.")

                self.logger.info("STEP: Verify that the folder was re-used.")
                self.assertEqual(
                    first_stat,
                    directory.stat(),
                    "Second directory is not the same as the first directory.",
                )

    def test_test_directory_function_call(self):
        """Test that the test directory method calls a method when created if supplied.

        Approval criteria:
            - The workspace library shall call method on directory creation if supplied.
            - The workspace library shall not call method if directory was not created.

        Test steps::
            1. Initialize the workspace.
            2. Enter a test directory with a method call registered.
            3. Verify that the method was called.
            4. Enter a test directory with the same identifer and a method call registered.
            5. Verify that the method was not called.
        """
        calls = []

        def callee(calls):
            """Test function for counting calls."""
            calls.append(1)

        self.logger.info("STEP: Initialize the workspace.")
        with Workspace(Mock()) as workspace:
            self.workspace = workspace

            self.logger.info(
                "STEP: Enter a test directory with a method call registered."
            )
            with workspace.test_directory("dir1", callee, calls):
                self.logger.info("STEP: Verify that the method was called.")
                self.assertEqual(len(calls), 1)

            self.logger.info(
                "STEP: Enter a test directory with the same identifer and a "
                "method call registered."
            )
            with workspace.test_directory("dir1", callee, calls):
                self.logger.info("STEP: Verify that the method was not called.")
                self.assertEqual(len(calls), 1)

    def test_test_directory_no_workspace(self):
        """Test that test directory raises exception if not workspace exist.

        Approval criteria:
            - The workspace library shall raise an exception on test directory if workspace
              does not exist.

        Test steps::
            1. Enter a test directory without a workspace.
            2. Verify that an exception was raised.
        """
        self.logger.info("STEP: Enter a test directory without a workspace.")
        self.workspace = Workspace(Mock())
        self.logger.info("STEP: Verify that an exception was raised.")
        with self.assertRaises(Exception):
            with self.workspace.test_directory("dir1"):
                pass



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests full executions."""
import os
import logging
from copy import deepcopy
from shutil import rmtree
from contextlib import contextmanager
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch
from etos_lib.lib.debug import Debug
from etos_test_runner.etr import ETR


SUITE = {
    "name": "Test ETOS API scenario",
    "priority": 1,
    "recipes": [
        {
            "constraints": [
                {"key": "ENVIRONMENT", "value": {}},
                {"key": "PARAMETERS", "value": {}},
                {"key": "COMMAND", "value": "exit 0"},
                {
                    "key": "TEST_RUNNER",
                    "value": "registry.nordix.org/eiffel/etos-python-test-runner:3.9.0",
                },
                {"key": "EXECUTE", "value": ["echo 'this is the pre-execution step'"]},
                {
                    "key": "CHECKOUT",
                    "value": [
                        "git clone https://github.com/eiffel-community/etos-test-runner.git ."
                    ],
                },
            ],
            "id": "6e8d29eb-4b05-4f5e-9207-0c94438479c7",
            "testCase": {
                "id": "ETOS API functests",
                "tracker": "Github",
                "url": "https://github.com/eiffel-community/etos-api",
            },
        }
    ],
    "test_runner": "registry.nordix.org/eiffel/etos-python-test-runner:3.9.0",
    "iut": {
        "provider_id": "default",
        "identity": "pkg:docker/production/etos/etos-api@1.2.0",
        "type": "docker",
        "namespace": "production/etos",
        "name": "etos-api",
        "version": "1.2.0",
        "qualifiers": {},
        "subpath": None,
    },
    "artifact": "e9b0c120-8638-4c73-9b5c-e72226415ae6",
    "context": "fde87097-46bd-4916-b69f-48dbbec47936",
    "executor": {},
    "log_area": {
        "provider_id": "default",
        "livelogs": "http://localhost/livelogs",
        "upload": {"url": "http://localhost/logs", "method": "POST"},
        "logs": {},
    },
}


class TestFullExecution(TestCase):
    """Test a full execution of ETR."""

    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(cls):
        """Create a debug instance."""
        cls.debug = Debug()

    def _patch_wait_for_request(self):
        """Patch the ETOS library wait for request method."""
        patcher = patch("etos_lib.etos.Http.wait_for_request")
        self.patchers.append(patcher)
        self.wait_for_request = patcher.start()
        self.suite = deepcopy(SUITE)
        self.wait_for_request.return_value = [self.suite]

    def _patch_http_request(self):
        """Patch the ETOS library http request method."""
        patcher = patch("etos_lib.etos.Http.request")
        self.patchers.append(patcher)
        self.http_request = patcher.start()
        self.http_request.return_value = True

    def _patch_test_suite_started_request(self):
        """Patch the GraphQL request for test suite started."""
        patcher = patch("etos_test_runner.lib.testrunner.request_test_suite_started")
        self.patchers.append(patcher)
        self.request_test_suite_started = patcher.start()
        self.request_test_suite_started.return_value = [
            {
                "data": {"testSuiteCategories": []},
                "meta": {"id": "577381ad-8356-4939-ab77-02e7abe06699"},
            }
        ]

    def setUp(self):
        """Create patch objects and a test folder to execute from."""
        self.patchers = []
        self.original = Path.cwd()
        self.root = Path().joinpath("testfolder").absolute()
        if self.root.exists():
            rmtree(self.root)
        self.root.mkdir()
        os.chdir(self.root)
        self._patch_wait_for_request()
        self._patch_http_request()
        self._patch_test_suite_started_request()

    def tearDown(self):
        """Clear the test folder and patchers."""
        os.chdir(self.original)
        rmtree(self.root, ignore_errors=True)
        for patcher in self.patchers:
            patcher.stop()

    @staticmethod
    @contextmanager
    def environ(add_environment):
        """Set environment variables in context and remove after.

        :param add_environment: Environment variables to add.
        :type add_environment: dict
        """
        current = {}
        for key, value in add_environment.items():
            current[key] = os.getenv(key)
            os.environ[key] = str(value)
        yield
        for key, value in current.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value

    def validate_event_name_order(self, events):
        """Validate ETR sent events.

        :raises AssertionError: If events are not correct.

        :param events: All events sent, in order.
        :type events: deque
        """
        self.logger.info(events)
        event_names_in_order = [
            "EiffelActivityTriggeredEvent",
            "EiffelActivityStartedEvent",
            "EiffelTestSuiteStartedEvent",
            "EiffelArtifactCreatedEvent",
            "EiffelArtifactPublishedEvent",
            "EiffelTestSuiteFinishedEvent",
            "EiffelConfidenceLevelModifiedEvent",
            "EiffelActivityFinishedEvent",
        ]
        for event_name in event_names_in_order:
            self.assertEqual(events.popleft().meta.type, event_name)
        self.assertEqual(list(events), [])

    def test_full(self):
        """Test that a full execution scenario works as expected.

        Approval criteria:
            - It shall be possible to execute a full suite in ETR.
            - ETR shall send events in the correct order.

        Test steps::
            1. Initialize and run ETR.
            2. Verify that events were sent in the correct order.
            3. Verify that ETR returned with status code 0.
        """

        environment = {
            "ETOS_DISABLE_SENDING_EVENTS": "1",
            "ETOS_DISABLE_RECEIVING_EVENTS": "1",
            "ETOS_GRAPHQL_SERVER": "http://localhost/graphql",
            "SUB_SUITE_URL": "http://localhost/download_suite",
            "HOME": self.root,  # There is something weird with tox and HOME. This fixes it.
        }
        with self.environ(environment):
            self.logger.info("STEP: Initialize and run ETR.")
            etr = ETR()
            result = etr.run_etr()

            self.logger.info("STEP: Verify that events were sent in the correct order.")
            self.validate_event_name_order(self.debug.events_published.copy())

            self.logger.info("STEP: Verify that ETR returned with status code 0.")
            # Result is either dictionary with outcome or an exit status code.
            # Exit status code on success is 0
            self.assertEqual(result, 0)
